# nexus-client

## Introduction

**nexus-client** is the client component of the Nexus VDI (Virtual Desktop Infrastructure) project.

## Building nexus-clients and nexus-exam

To build `nexush`, `nexus-cli` and `nexus-exam` read [INSTALL.md](INSTALL.md).

## Obtaining nexus-clients

Precompiled nexus client binaries can be [found here](https://drive.switch.ch/index.php/s/Llr4w51smZitJ7u).

Two clients are available:

- `nexush` is the most "user-friendly" client; it is the nexus shell
- `nexus-cli` is another client more suited to scripting

Both clients feature the same commands and they are available for multiples operating systems and processor architectures:

| Operating System | Architecture  | Clients directory location |
| ---------------- | ------------- | -------------------------- |
| Linux            | x86-64/AMD64  | binaries/amd64/linux       |
| Linux            | Intel 32-bits | binaries/386/linux         |
| Windows          | x86-64/AMD64  | binaries/amd64/windows     |
| Windows          | Intel 32-bits | binaries/386/windows       |
| OSX              | x86-64/AMD64  | binaries/amd64/darwin      |
| Linux            | ARM64         | binaries/arm64/linux       |
| Windows          | ARM64         | binaries/arm64/windows     |
| OSX              | ARM64         | binaries/arm64/darwin      |

## Installing nexus-clients' dependency

Both clients require the `remote-viewer` software in order to display remote VM desktops.

### Installing remote-viewer on Linux

On Ubuntu/Debian, `remote-viewer` is part of the `virt-viewer` package which can be installed with:
```
sudo apt-get install virt-viewer
```

### Installing remote-viewer on Windows

On Windows, `remote-viewer` is part of the `virt-viewer` MSI package that can be downloaded here: [https://virt-manager.org/download.html](https://virt-manager.org/download.html).

Once installed, make sure to **add the folder** where `remote-viewer.exe` is installed to your user's **PATH environment variable** (or the system PATH).
Then reboot your system.

### Installing remote-viewer on Mac OSX

On Mac, `remote-viewer` is part of the `virt-viewer` package from MacPorts that can be installed from here [https://ports.macports.org/port/virt-viewer/](https://ports.macports.org/port/virt-viewer/).

## Overview of nexus' basic concepts

### Templates

- Templates are immutable OS disk images used to create VMs: a VM is an instance of a specific template
- Templates can be `private` or `public`: a private template is only visible to its owner whereas a public template is visible to everyone
- **Important**: templates don't store any hardware configuration (number of CPUs, amount of RAM, network interface, etc.); they only store the disk image, nothing more

### VMs

- A VM is always created from a template
- A VM's hardware (number of CPUs, amount of RAM, network interface, etc.) is completely independant of its disk content
- Access rights to a VM can be finely controlled by its owner and anyone else allowed (see [Access control](#access-control))

### Users

- What users can or cannot do is defined by their capabilities: see [Access control](#access-control) for more information

## Available template images

The table below describes the basic public templates that are currently available. The numbers between parenthesis indicate the year and month of creation (also available in the detailed output of `tpllist -l <template ID>`).

| Template name                                         | Description                                                         |
| ----------------------------------------------------- | ------------------------------------------------------------------- |
| Xubuntu 22.04 (2024.01) + dev env/unpriv user         | Xubuntu 22.04 system (Ubuntu with XFCE desktop environment)         |
|                                                       | + C dev tools (gcc, make, etc.), VScode IDE compiler                |
|                                                       | + QEMU system hypervisor                                            |
|                                                       | `nexus` user (`/home/nexus`) with sudo privileges (pwd `nexus`)     |
|                                                       | `student` user (`/home/student`) without any privilege (no pwd)     |
| Xubuntu 22.04 (2024.01) + dev env/unpriv user/pi-hole | Xubuntu 22.04 system (Ubuntu with XFCE desktop environment)         |
|                                                       | + C dev tools (gcc, make, etc.), VScode IDE compiler                |
|                                                       | + QEMU system hypervisor                                            |
|                                                       | + Pi-hole domain firewall (see [Domain firewall](#domain-firewall)) |
|                                                       | `nexus` user (`/home/nexus`) with sudo privileges (pwd `nexus`)     |
|                                                       | `student` user (`/home/student`) without any privilege (no pwd)     |
| Manjaro 21 xfce (2022.08)                             | Vanilla Manjaro 21 system with XFCE desktop environment             |
|                                                       | `nexus` user (`/home/nexus`) with sudo privileges (pwd `nexus`)     |

<!--
| Xubuntu 22.04 (2022.08) + dev env                     | Xubuntu 22.04 system (Ubuntu with XFCE desktop environment)     |
|                                                       | + C dev tools (gcc, make, etc.), VScode IDE compiler            |
|                                                       | + QEMU system hypervisor                                        |
|                                                       | `nexus` user (`/home/nexus`) with sudo privileges (pwd `nexus`) |
| Debian 11 xfce (2022.08)                              | Vanilla Debian 11 system with XFCE desktop environment          |
|                                                       | `nexus` user (`/home/nexus`) with sudo privileges (pwd `nexus`) |
| Ubuntu 22.04 (2022.08)                                | Vanilla Ubuntu 22.04 system with the GNOME desktop environment  |
|                                                       | `nexus` user (`/home/nexus`) with sudo privileges (pwd `nexus`) |
-->

## nexush

`nexush` is a **nexus-client** in the form of a single native executable featuring the commands listed below.
Most commands support regular expressions (regex) in order to perform actions on multiple VMs at once.

Regular expressions must conform to the RE2 GO syntax described here [https://github.com/google/re2/wiki/Syntax](https://github.com/google/re2/wiki/Syntax).

`nexush` features many commands to manipulate:

- VMs (commands starting with `vm`)
- templates (commands starting with `tpl`)
- users (commands starting with `user`)

It also supports command completion through the `tab` key (like `bash`).

Below, is an example of execution showing all available commands:

```
Welcome to nexush, the nexus shell.
Type: "help" for help on commands
      "quit" or "exit" to quit nexush
nexush> help
═════╡ GENERAL commands ╞═════════════════════════════════════════════════════════════════

ls                  List files in the specified dir or in the current dir if no argument is specified.

refresh             Obtains a new access token.

version             Display nexus server and client's versions.

═════╡ USER commands ╞═════════════════════════════════════════════════════════════════

whoami              Displays the current user's details.

passwd              Updates the current user's password.

userlist            Lists users.
                    Requires USER_LIST user capability.

usercreate          Creates a user.
                    Requires USER_CREATE user capability.

userdel             Deletes one or more users.
                    Requires USER_DESTROY user capability.

userresetpwd        Reset password for one or more users.
                    Requires USER_RESETPWD user capability.

usersetcaps         Sets a user's capabilities.
                    Requires USER_SET_CAPS user capability.

userunlock          Unlocks one or more users.
                    Requires USER_UNLOCK user capability.

═════╡ TEMPLATE commands ╞═════════════════════════════════════════════════════════════════

tplcreate           Creates a template from an existing VM.
                    Requires TPL_CREATE user capability.

tpldel              Deletes a template.
                    Requires TPL_DESTROY or TPL_DESTROY_ANY user capability.

tpledit             Edits one or more template's properties: name, access.
                    Requires TPL_EDIT or TPL_EDIT_ANY user capability.

tplexportdisk       Exports a template's disk.
                    Requires TPL_READFS or TPL_READFS_ANY user capability.

tpllist             Lists templates.
                    Requires TPL_LIST or TPL_LIST_ANY user capability.

═════╡ VM commands ╞═════════════════════════════════════════════════════════════════

vmaddaccess         Adds a user's VM access in one or more VMs.
                    If not the VM's owner: requires VM_SET_ACCESS user capability and VM_SET_ACCESS VM access capability.

vmattach            Attaches to one or more VMs in order to use their desktop environment.
                    If not the VM's owner: requires VM_ATTACH VM access capability or VM_ATTACH_ANY user capability.

vmcreate            Creates one or more VMs.
                    Requires VM_CREATE user capability.

vmcreds2pdf         Creates a PDF with the credentials required to attach to running VMs.
                    If not the VM's owner: requires VM_ATTACH VM access capability or VM_ATTACH_ANY user capability.

vmcreds2csv         Creates a CSV file with the credentials required to attach to running VMs.
                    The written CSV file contains 3 columns: VM ID;VM name;password
                    If not the VM's owner: requires VM_ATTACH VM access capability or VM_ATTACH_ANY user capability.

vmdel               Deletes one or more VMs.
                    If not the VM's owner: requires VM_DESTROY VM access capability or VM_DESTROY_ANY user capability.

vmdelaccess         Removes a user's VM access in one or more VMs.
                    If not the VM's owner: requires VM_SET_ACCESS user capability and VM_SET_ACCESS VM access capability.

vmedit              Edits one or more VMs' properties: name, cpus, ram or nic.
                    If not the VM's owner: requires VM_EDIT VM access capability or VM_EDIT_ANY user capability.

vmexportdir         Exports one or more VMs' directory into one or more compressed archives.
                    If not the VM's owner: requires VM_READFS VM access capability or VM_READFS_ANY user capability.

vmimportdir         Copies a local directory (or file) and all its content into one or more VMs.
                    If not the VM's owner: requires VM_WRITEFS VM access capability or VM_WRITEFS_ANY user capability.

vmkill              Kills one or more VMs.
                    If not the VM's owner: requires VM_STOP VM access capability or VM_STOP_ANY user capability.

vmlist              Lists VMs.
                    If not the VM's owner: requires VM_LIST VM access capability or VM_LIST_ANY user capability.

vmreboot            Gracefully reboots one or more VMs.
                    If not the VM's owner: requires VM_REBOOT VM access capability or VM_REBOOT_ANY user capability.

vmshutdown          Gracefully shutdowns one or more VMs.
                    If not the VM's owner: requires VM_STOP VM access capability or VM_STOP_ANY user capability.

vmstart             Starts one or more VMs.
                    If not the VM's owner: requires VM_START VM access capability or VM_START_ANY user capability.

vmstartwithcreds    Starts one or more VMs with user-defined credentials.
                    If not the VM's owner: requires VM_START VM access capability or VM_START_ANY user capability.
```

Note that when attached to a VM's desktop (`vmattach` command), ctrl+F12 toggles between fullscreen/non-fullscreen modes.

## nexus-cli

`nexus-cli` is another **nexus-client** in the form of a single native executable featuring pretty much the exact same commands as `nexush`. Therefore we won't list them here.

## USB redirection

USB redirection allows USB devices plugged in the PC running nexus client to be accessible in the guest OS of the VM running on the server.

When creating a new VM with `vmcreate`, you can specify USB devices that are accessible in the VM. A USB device is specified by two hexadecimal number, a vendor ID (`VID`) and a product ID (`PID`). For instance, the example below creates a VM and allows two USB devices, 067b:2303 (a USB/serial connector) and 0781:5567 (a SanDisk USB key) to be accessible in the VM:
```
vmcreate "usb test" 2 4000 none 067b:2303,0781:5567 fbccb584-9ea6-40f7-926d-dabf3970525e
```

If you wish to forbid all USB devices simply specify "none" instead:
```
vmcreate "usb test" 2 4000 none none fbccb584-9ea6-40f7-926d-dabf3970525e
```

To obtain the vendor ID and product ID of a USB device, run the `lsusb` command on Linux. The output should be similar to the one below, where the sixth column indicates vendor ID and product ID separated by a colon (`:`):
```
Bus 003 Device 026: ID 04d8:0b27 Microchip Technology, Inc. USB2734
Bus 003 Device 003: ID 0b0e:245d GN Netcom Jabra Link 370
Bus 003 Device 009: ID 058f:6362 Alcor Micro Corp. Flash Card Reader/Writer
Bus 003 Device 006: ID 046d:08e3 Logitech, Inc. C505 HD Webcam
```

### Vendor ID and product ID of USB devices of interest

| Device description                           | VID:PID   |
| -------------------------------------------- | --------- |
| Altera USB-Blaster                           | 09fb:6001 |
| Atmel at91sam SAMBA bootloader               | 03eb:6124 |
| Digilent Analog Discovery 2                  | 0403:6014 |
| MyLab2 serial port                           | 0403:6015 |
| NXP Semiconductors LPC11U3x CMSIS-DAP v1.0.4 | 1fc9:0132 |
| NXP Semiconductors NXP CMSIS-DAP             | 1fc9:001d |
| Peak USB CAN                                 | 0c72:000c |
| Prolific Technology USB/serial               | 067b:2303 |
| Saleae Logic Pro 16                          | 21a9:1006 |
| SanDisk Cruzer Blade USB key                 | 0781:5567 |
| Segger J-Link                                | 1366:0101 |
| USBest Technology USB key                    | 1307:0165 |
| Xilinx KV260 and KR260 JTAG and UART         | 0403:6011 |
| Xilinx Nexys Video JTAG                      | 0403:6010 |
| Xilinx Nexys Video UART                      | 0403:6001 |

## Tutorial: creating a live exam with nexush

The VM templates available on nexus server are described here: [available template images](#available-template-images).

Let's say you want to create an exam for the class "ProgSys". Let's assume 30 students are enrolled in the class.

First, you need to create the VM that will be used by your students during the exam. Typically, this VM will contain the exam environment (compilers, editors, tools, etc.) along the description of the exam, for instance as a PDF on the desktop.

1. First, log-in with your user:
   ```
   ./nexush janedoe@nexus.org
   ```
   After entering your password, you should see:
   ```
   janedoe@nexus.org's password:
   Welcome to nexush, the nexus shell.
   Type: "help" for help on commands
         "quit" or "exit" to quit nexush
   nexush>
   ```

1. List available templates and choose the one you wish to use for your exam VM:
   ```
   tpllist .
   ```
   This command displays the templates you can use:
   ```
   Manjaro 21 xfce (2022.08)                                  | 502c30cb-49ea-4a83-a405-8c9182d9970c | public
   Xubuntu 22.04 (2024.01) + dev env/unpriv user              | 6d735b5d-1ff7-4931-b6f7-0b8eb070d7bd | public
   Xubuntu 22.04 (2024.01) + dev env/unpriv user/pi-hole      | 17152441-7199-4b0a-8570-2ade426b4946 | public
   ```
   Let's assume template `6d735b5d-1ff7-4931-b6f7-0b8eb070d7bd` (Xubuntu 22.04 (2024.01) + dev env/unpriv user), is the one you would like to use.
   As a reminder, this template's OS features two users:

   - student: unprivileged user (doesn't have a sudo access); its password is "student"
   - nexus: privileged user  (access to sudo); its password is "nexus"

1. Create the VM based on the chosen template. Let's say you want the VM to be named "Exam ProgSys 06.2024" and you want it to have 2 CPUs, 3GB RAM and access to the Internet (for now):
   ```
   vmcreate "Exam ProgSys 06.2024" 2 3000 user none 6d735b5d-1ff7-4931-b6f7-0b8eb070d7bd
   ```
   This command displays the name and ID of the created VM(s):
   ```
   Created VM "Exam ProgSys 06.2024" | 148ea20e-9c45-4bd1-a670-f18312d436b8
   ```

1. Now that the VM is created, you need to start it, connect to it, and configure it to fit your needs. First, start it by starting any VM matching the following expression (matching is case-insensitive - you can also start it by specifying its unique VM ID):
   ```
   vmstart "exam progsys 06.2024"
   ```
   This command displays the name of the started VM(s):
   ```
   Started VM "Exam ProgSys 06.2024"
   ```

1. Attach to the VM in order to configure it to your needs:
   ```
   vmattach "exam progsys 06.2024"
   ```
   alternatively, you could use the VM ID instead (this ensure any other possible name matches):
   ```
   vmattach 148ea20e-9c45-4bd1-a670-f18312d436b8
   ```
   This command opens a window showing the VM(s)'s desktop:
   ![](img/vmattach.jpg)

1. Configure the VM to suit your needs, typically by:
   - switching to user "nexus"
     ```
     su nexus
     ```
   - updating the system (not required):
     ```
     su nexus
     sudo apt-get update && apt-get upgrade -y
     ```
   - installing applications required for the exam, e.g. compilers, tools, editors, etc.:
     ```
     sudo apt-get install -y ...
     ```
   - configuring the desktop to suit your needs and preferences, emptying the trash, etc.
   <!--
   - installing the QEMU Guest Agent with `sudo apt-get install qemu-guest-agent` (so you will be able to use the `vmshutdown` command)
   -->

1. Add the exam' specific files to your VM. On your local system, simply create an archive (such as `.tar.gz` or `.zip`) of the files and directories you want to copy to your VM. Once completed, simply drag-and-drop it to your VM's desktop window. This will transfer the archive into your VM's desktop directory. Then, in the VM, simply decompress it by clicking on it with the right button and selecting "Extract Here".

<!--
1. Add the exam' specific files. All files (questions, code, etc.) can be copied from your local machine to the VM with `vmimportdir`. However, this command requires the VM to be shutdown so make sure it's not running. Below, we copy the "exam" local directory into the VM's student user's Desktop (`/home/student/Desktop`). Note that it copies the directory along all its subdirectories and files:
   ```
   vmimportdir "exam progsys 06.2024" exam /home/student/Desktop
   ```
   The command displays the directory was copied successfully:
   ```
   Successfully copied "exam" into "/home/student/Desktop" in VM "Exam ProgSys 06.2024"
   ```
-->

1. Now that the VM is ready for the exam, create a template from it. The template must be `private` as we don't want anyone else to access it. Let's choose "Exam ProgSys 06.2024" as the template name (template creation takes several minutes, the larger the VM, the longer). The first argument is the template's name and the last is the VM ID from which the template must be created (as displayed when the VM was created earlier):
   ```
   tplcreate "Exam ProgSys 06.2024" private 148ea20e-9c45-4bd1-a670-f18312d436b8
   ```
   This command displays details about the newly created template:
   ```
   {
       "id": "0fb0b1f2-c72c-416e-961a-6bb802da89bb",
       "name": "Exam ProgSys 06.2024",
       "owner": "janedoe@nexus.org",
       "access": "private"
       "creationTime": "2024-06-15T16:42:17+01:00"
   }
   ```

1. You can now create the VMs for your 30 students using your new template (ID `0fb0b1f2-c72c-416e-961a-6bb802da89bb`). Let's say the base name for the 30 VMs to create is "Live Exam ProgSys 06.2024". You want the VMs to have: 2 CPUs, 3GB RAM and no network interface to forbid any potentiel network/Internet access. Create a CSV file, say `students.csv` with the names of your 30 students, one per line (single column file). Then, create the 30 VMs with:
   ```
   vmcreate "Live Exam ProgSys 06.2024" 2 3000 none none 0fb0b1f2-c72c-416e-961a-6bb802da89bb students.csv
   ```
   This command displays the list of created VMs:
   ```
   Created VM "Live Exam ProgSys 06.2024 <Alia Friedman>" | 74d8b83d-f59e-4129-bf68-af574968cf48
   Created VM "Live Exam ProgSys 06.2024 <Aria Doyle>" | f3047faa-2f15-4f47-b79f-9acc19751b6c
   Created VM "Live Exam ProgSys 06.2024 <Avah Coffey>" | 3ebd56a2-2c1e-416c-9847-f80ee3efa1c1
   Created VM "Live Exam ProgSys 06.2024 <Briley Brady>" | 245fc5b2-b192-4b41-80be-2d39b5a2cef2
   Created VM "Live Exam ProgSys 06.2024 <Brooklyn Sweeney>" | a9bafd7e-28f0-4f37-8b90-5e3c82d4bbc5
   Created VM "Live Exam ProgSys 06.2024 <Cornelius Simmons>" | 40edb2b1-b4e9-4928-9dea-316ed834bf07
   Created VM "Live Exam ProgSys 06.2024 <Donovan Heath>" | 8ae5c9cd-16f3-4e02-a940-e4209a6d7010
   Created VM "Live Exam ProgSys 06.2024 <Ella Webster>" | 7d16f88f-afb1-4633-a646-57a9c87411d5
   ...
   ```

1. The day of the exam, you'll have to start the 30 VMs and generate a PDF with the randomly generated credentials required to connect to each VM (a password, unique to each VM). To start the 30 VMs, run:
   ```
   vmstart "live exam progsys"
   ```
   This command displays the list of started VMs:
   ```
   Started VM "Live Exam ProgSys 06.2024 <Alia Friedman>"
   Started VM "Live Exam ProgSys 06.2024 <Aria Doyle>"
   Started VM "Live Exam ProgSys 06.2024 <Avah Coffey>"
   Started VM "Live Exam ProgSys 06.2024 <Briley Brady>"
   Started VM "Live Exam ProgSys 06.2024 <Brooklyn Sweeney>"
   Started VM "Live Exam ProgSys 06.2024 <Cornelius Simmons>"
   Started VM "Live Exam ProgSys 06.2024 <Donovan Heath>"
   Started VM "Live Exam ProgSys 06.2024 <Ella Webster>"
   ...
   ```

1. Finally, to produce the PDF file, `creds.pdf`, with all credentials required to attach to each VM, run:
   ```
   vmcreds2pdf "live exam progsys" creds.pdf
   ```
   This command displays that file was written successfully:
   ```
   creds.pdf written successfully.
   ```
   Here is what the PDF looks like:
   
   ![](img/vmcred2pdf.png)

1. It is strongly recommended to save the VMs' credentials in a CSV file as it's possible that some students might shutdown their VM or that a VM would crash (rare, but possible). In this case, restarting the VMs would generate new credentials, different from the printed ones. We recommend you to save the VMs credentials in a CSV file with `vmcreds2csv "live exam progsys" creds.csv` so that later, the VMs can be restarted with the exact same credentials using `vmstartwithcreds creds.csv`.

1. You can now print the PDF above and cut each line, one per student.

1. The day of the exam:
   - Give each student their strip of paper so that they can log into their VM.
   - Once a student has completed their exam, they should shutdown their VM and give you the strip of paper back. Note that you can also shutdown the VM yourself with `vmshutdown`. You can also force a VM to be stopped with `vmkill`, but it's discouraged as it might lead to filesystem corruptions as the VM is abruptly killed.
   - You can also shutdown all VMs at a given time with `vmshutdown "live exam progsys"`

1. To correct the students' exams, simply download the relevant files to your computer using `vmexportdir`. As an example, the command below extracts and downloads the `/home/student/Desktop/exam` directory of all VMs matching "Live Exam ProgSys" (each directory is saved in a `.tar.gz` archive named after the VM's name):
   ```
   vmexportdir "live exam progsys" /home/student/Desktop/exam
   ```
   This command displays each exported file tree:
   ```
   Successfully exported /home/student/Desktop/exam from VM "Live Exam ProgSys 06.2024 <Alia Friedman>" into Live Exam ProgSys 06.2024 <Alia Friedman>.tar.gz
   Successfully exported /home/student/Desktop/exam from VM "Live Exam ProgSys 06.2024 <Aria Doyle>" into Live Exam ProgSys 06.2024 <Aria Doyle>.tar.gz
   Successfully exported /home/student/Desktop/exam from VM "Live Exam ProgSys 06.2024 <Avah Coffey>" into Live Exam ProgSys 06.2024 <Avah Coffey>.tar.gz
   Successfully exported /home/student/Desktop/exam from VM "Live Exam ProgSys 06.2024 <Briley Brady>" into Live Exam ProgSys 06.2024 <Briley Brady>.tar.gz
   Successfully exported /home/student/Desktop/exam from VM "Live Exam ProgSys 06.2024 <Brooklyn Sweeney>" into Live Exam ProgSys 06.2024 <Brooklyn Sweeney>.tar.gz
   Successfully exported /home/student/Desktop/exam from VM "Live Exam ProgSys 06.2024 <Cornelius Simmons>" into Live Exam ProgSys 06.2024 <Cornelius Simmons>.tar.gz
   Successfully exported /home/student/Desktop/exam from VM "Live Exam ProgSys 06.2024 <Donovan Heath>" into Live Exam ProgSys 06.2024 <Donovan Heath>.tar.gz
   Successfully exported /home/student/Desktop/exam from VM "Live Exam ProgSys 06.2024 <Ella Webster>" into Live Exam ProgSys 06.2024 <Ella Webster>.tar.gz
   ...
   ```

## nexush usage examples

Launch `nexush` and log in as user `janedoe@nexus.org` (note that you will be prompted for your password):
```
./nexush janedoe@nexus.org
```

Display available commands:
```
help
```

Display the help for the `vmcreate` command:
```
vmcreate
```

Display information about the currently logged-on user:
```
whoami
```

List all users:
```
userlist .
```

List users matching the "jane" pattern:
```
userlist jane
```

Add new user `lukeskywalker@force.net` with a list of capabilities:
```
useradd lukeskywalker@force.org Luke Skywalker maytheforcebewithyou USER_CREATE USER_DESTROY USER_LIST USER_SET_CAPS VM_CREATE
```

List all listable VMs:
```
vmlist .
```

List all listable VMs with more details ("long output"):
```
vmlist -l .
```

List listable VMs matching the "ubuntu" pattern:
```
vmlist ubuntu
```

List listable VMs matching the "ubuntu" pattern and also the VM with ID `6713ce26-941e-4d95-8e92-6b71d44bf75a`:
```
vmlist ubuntu 6713ce26-941e-4d95-8e92-6b71d44bf75a
```

Start VM `6713ce26-941e-4d95-8e92-6b71d44bf75a`:
```
vmstart 6713ce26-941e-4d95-8e92-6b71d44bf75a
```

Start VMs matching the "exam ISC_433 PCO" pattern:
```
vmstart "exam ISC_433 PCO"
```

Attach to VM `6713ce26-941e-4d95-8e92-6b71d44bf75a` and all VMs matching the pattern "zorglub":
```
vmattach 6713ce26-941e-4d95-8e92-6b71d44bf75a zorglub
```

Shutdown VMs matching the "exam ISC_433 PCO" pattern (for this to work, `qemu-guest-agent` must be running in the VM's guest OS!):
```
vmshutdown "exam ISC_433 PCO"
```

Kill VM `6713ce26-941e-4d95-8e92-6b71d44bf75a`:
```
vmkill 6713ce26-941e-4d95-8e92-6b71d44bf75a
```

Create a VM named "Doom", based on the `fbccb584-9ea6-40f7-926d-dabf3970525e` (Doom) template, with 4 CPUs, 4GB RAM, a network interface with NAT translation, and no USB devices:
```
vmcreate Doom 4 4096 user none fbccb584-9ea6-40f7-926d-dabf3970525e
```

Create 50 VMs with the base name "ISC_433 Exam" based on the `6713ce26-941e-4d95-8e92-6b71d44bf75a` template, with 2 CPUs, 2GB RAM, no network interface, and no USB devices:
```
vmcreate "ISC_433 Exam" 2 2048 none none 6713ce26-941e-4d95-8e92-6b71d44bf75a 50
```
It takes about 30 seconds and 11MB of disk space to create these 50 VMs.
They will have the following names:
```
ISC_433 Exam <1>
ISC_433 Exam <2>
...
ISC_433 Exam <50>
```

Edit VM `6713ce26-941e-4d95-8e92-6b71d44bf75a` by changing its name to "Tagada VM" and no network interface (`none`):
```
vmedit 6713ce26-941e-4d95-8e92-6b71d44bf75a name="Tagada VM" nic=none
```

Edit VMs matching the "PCO lab2" pattern by changing their CPU to 1 core, a network interface with NAT translation (`user`), and adding access to USB devices baba:0007 and 1234:abcd:
```
vmedit "PCO lab2" cpus=1 nic=user usb=baba:0007,1234:abcd
```

Delete VM `6713ce26-941e-4d95-8e92-6b71d44bf75a`:
```
vmdel 6713ce26-941e-4d95-8e92-6b71d44bf75a
```

Delete VMs matching the "exam ISC_433 PCO" pattern:
```
vmdel "exam ISC_433 PCO"
```

Set the VM access for VM `89649fe3-4940-4b77-929e-50903789cd87` with: `VM_LIST` and `VM_ATTACH` for user `student@nexus.org`:
```
vmaddaccess 89649fe3-4940-4b77-929e-50903789cd87 student@nexus.org VM_LIST VM_ATTACH
```

Set VM access for VMs matching the "alpine" pattern with: `VM_START` and `VM_STOP` for user `student@nexus.org`:
```
vmaddaccess alpine student@nexus.org VM_START VM_STOP
```

Remove VM access for `student@nexus.org` from VM `89649fe3-4940-4b77-929e-50903789cd87`:
```
vmdelaccess 89649fe3-4940-4b77-929e-50903789cd87 student@nexus.org
```

Remove VM access for `student@nexus.org` from VMs matching the "lab2" pattern:
```
vmdelaccess lab2 student@nexus.org
```

Generate `exam_vms.pdf` with the credentials required to connect to all running VMs matching "exam prog sys":
```
vmcreds2pdf "exam prog sys" output.pdf
```

Extract and download the `/home` directory of all VMs matching "exam prog sys" (each directory is saved in a `.tar.gz` archive named after the VM's ID):
```
vmexportdir "exam prog sys" /home
```

List all available templates:
```
tpllist .
```

List templates matching the "ubuntu" pattern:
```
tpllist ubuntu
```

Create a new `public` template, named "Xubuntu 22.04 + golang toolchain" based on VM `89649fe3-4940-4b77-929e-50903789cd87` (`public` templates are accessible to everyone while `private` templates are only accessible to their creators):
```
tpllist 89649fe3-4940-4b77-929e-50903789cd87 "Xubuntu 22.04 + golang toolchain" public
```

Delete template `3d440a31-17da-423d-8d95-a96b4cecff8b`:
```
tpldel 3d440a31-17da-423d-8d95-a96b4cecff8b
```

Quit `nexush`:
```
quit
```

## nexus-cli usage examples

Given `nexus-cli` offers the same VM, user and template commands as `nexush`, we only show a handful of examples here.

List available commands:
```
./nexus-cli
```

Authentify user `janedoe@nexus.org` and obtain an access token:
```
export NEXUS_TOKEN=`./nexus-cli login janedoe@nexus.org pipomolo`
```

Start VM `6713ce26-941e-4d95-8e92-6b71d44bf75a`:
```
vmstart 6713ce26-941e-4d95-8e92-6b71d44bf75a
```

## nexus-exam

### Running nexus-exam

The purpose of `nexus-exam` is to be used during live exam sessions. Students are provided with laptops configured to boot on dedicated USB keys. These USB keys feature a minimal Linux system that automatically run `nexus-exam` at boot time. 80 such USB keys have been created and are ready to be used.
The only thing that's necessary to use `nexus-exam` is to boot on one of the USB keys mentionned above.

Additionally, `nexus-exam` can also be ran by booting on PXE on the nexus-exam image available to any loaned laptop connected to a RJ45 câble in a classroom or office.

Below is a screenshot of `nexus-exam`'s graphical interface:

![](img/nexus-exam.png)

<!--
## Advanced use: configuring nexush and nexus-cli

Because communication between nexus client and nexus server is encrypted, a public certificate is required. Both clients, `nexush` and `nexus-cli`, embed the necessary public certificate to communicate with the server. By default, they are also configured to communicate with the production nexus-server available at ISC.

Both clients can be configured to communicate with any nexus server. ISC has two nexus servers, one for production use and one for development. The corresponding public certificates can be found in the `certificates` directory found in [nexus-certs repository](https://gitedu.hesge.ch/flg_projects/nexus_vdi/nexus-certs).

Here are the ip addresses, ports and certificates to use ISC's production and development nexus servers:

| Server      | IP            | Port | Certificate                |
| ----------- | ------------- | ---- | -------------------------- |
| production  | 1.2.3.4       | 1077 | tagada.pem                 |
| development | 5.6.7.8       | 1077 | tsoin.pem                  |

Nexus clients use two environment variables to specify which nexus server to use and on which port, and which public certificate to use:

| Variable     | Description                                      |
| ------------ | ------------------------------------------------ |
| NEXUS_SERVER | server to use, using the syntax: `ip:port`       |
| NEXUS_CERT   | location (path) to the public certificate to use |

On Linux and Mac OSX, a simple and effective way is to store these variables in a `.env` file that is sourced before running the clients.

Here is an example of `.env.nexus` file configured to use the production ISC server:

```
export NEXUS_SERVER=127.0.0.1:1077
export NEXUS_CERT=$(pwd)/grmbl.pem
```

Then, source it and run your client of choice:
```
source .env.nexus
./nexush
```
-->

## Access control

Access control [is described here](access_control.md).

## REST API

The REST API [is described here](rest_api.md).

## Domain firewall

Some templates come with a domain firewall already installed: [Pi-hole](https://pi-hole.net/).

Pi-hole is a tool that lets you configure which domains are accessible and which are not. It can use whitelists or blacklists to define what's allowed or not and it's highly configurable through either a web interface or the command line.

To configure Pi-hole in the VM:

1. Log on with the `nexus` user as you'll need root privileges
1. Set the password to access the web interface by running the following in a terminal:
   ```
   pihole -a -p
   ```
1. Log on Pi-hole's interface by going to `http://pi.hole` in a browser
1. In the configuration, all domains are blocked except for the ones that are whitelisted; consequently, you must configure the domains you want to allow (you can also disable domain filtering entirely)
1. Once done, the settings will be persistent

The Pi-hole installation in the template(s) already come with domain rules to allow/forbid the three scenarios below:

- [Cyberlearn](https://cyberlearn.hes-so.ch)
- [Gitedu](https://gitedu.hesge.ch)
- Docker Hub (to allow `docker pull/run` from docker CLI)

## FAQ

### [1] What is QEMU Guest Agent?

QEMU Guest Agent is a service (a *daemon* in the UNIX terminology) that is meant to run inside a VM's guest OS. Its purpose is to interact with the hypervisor so that the hypervisor can read the state of the guest OS, send commands (e.g. shutdown, reset, etc.), read files, etc. Communications between hypervisor and QEMU Guest Agent is done through the QEMU Guest Agent Protocol. 

Certain nexus features require QEMU Guest Agent to run in the VM's guest OS. This is the case of the `vmshutdown` command which gracefully shutdowns a VM. This command won't work if QEMU Guest Agent is not running in the VM's guest OS.

All base OS VM templates provided in nexus feature QEMU Guest Agent, so users don't need to worry about it.

On a Ubuntu/Debian system, the `qemu-guest-agent` service can be installed with:
```
sudo apt-get install qemu-guest-agent
```

### [2] How to automatically change the display resolution when the window's dimensions change?

Some desktop environments automatically take care of this out of the box. Gnome in Ubuntu does it automatically, but XFCE doesn't do it for instance.

Below we desribe what to do to solve this issue. The idea is to execute a script that runs indefinitely. The scripts' purpose is to monitor the display output and whenever a change of resolution is detected, it adjusts the display to the newly detected resolution.

1. As root, create the bash script below and save it into `/usr/local/bin/update-virt-display`:
   ```bash
   #!/bin/bash

   # Indefinitely wait for a screen resolution change and when one happens, update the desktop size accordingly.
   xev -root -event randr | \
   grep --line-buffered 'subtype XRROutputChangeNotifyEvent' | \
   while read vars ; do \
      xrandr --output "$(xrandr | awk '/ connected/{print $1; exit; }')" --auto
   done
   ```
1. As root, make the script executable with:
   ```
   chmod +x /usr/local/bin/update-virt-display
   ```
1. Then, for each user you want to support, create a `.xprofile` file in the user's home directory with the following content:
   ```
   /usr/local/bin/update-virt-display &
   ```
