package utils

import (
    "os"
    "bufio"
    "strconv"
    "strings"
)

// This function returns info about the system's RAM
// by reading /proc/meminfo.
// Beware: it is Linux SPECIFIC!
// Returns total and avaiable memory in MB.
func GetRAM() (int, int, error) {
    f, err := os.Open("/proc/meminfo")
    if err != nil {
        return 0, 0, err
    }
    defer f.Close()

    scan := bufio.NewScanner(f)
    scan.Split(bufio.ScanLines)

    var memTotal int = -1
    var memAvail int = -1
    const units = 1024

    for scan.Scan() {
        s := scan.Text()
        fields := strings.Split(s, ":")
        attr := fields[0]
        vals := strings.Split(strings.TrimSpace(fields[1]), " ")
        val := strings.TrimSpace(vals[0])
        attr = strings.TrimSpace(strings.ToLower(attr))
        switch attr {
        case "memtotal":
            memTotal, _ = strconv.Atoi(val)
            memTotal = memTotal/units
        case "memavailable":
            memAvail, _ = strconv.Atoi(val)
            memAvail = memAvail/units
        }
    }

    return memTotal, memAvail, nil
}

/*
// This function is not used because it DOES NOT return the available RAM!
// Returns: total RAM, shared RAM, buffered RAM, free RAM and possibly an error.
// RAM amounts are given in megabytes (MB).
func GetRAM() (int, int, int, int, error) {
    in := &syscall.Sysinfo_t{}
    if err := syscall.Sysinfo(in); err != nil {
        return 0, 0, 0, 0, err
    }

    const MB = 1024*1024
    // On 32-bit systems these fields are uint32 instead of uint64.
    // We always convert to uint64 to match signature.
    totalRam :=  int(uint64(in.Totalram)*uint64(in.Unit)/MB)
    sharedRam :=  int(uint64(in.Sharedram)*uint64(in.Unit)/MB)
    bufferedRam := int(uint64(in.Bufferram)*uint64(in.Unit)/MB)
    freeRam := int(uint64(in.Freeram)*uint64(in.Unit)/MB)
    return totalRam, sharedRam, bufferedRam, freeRam, nil
}
*/